@@ -13,6 +13,7 @@ type flags = {
   defined_variables : StrSet.t;
   print_units : bool; (* debug *)
   time_typechecking : bool; (* bench type checker *)
+  skip_det_checking: bool;
 }
 val default_flags : flags
 
@@ -45,6 +46,7 @@ val query_of_scoped_term : program -> (State.t -> State.t * Compiler_data.Scoped
 val query_of_raw_term : program -> (State.t -> State.t * term * Conversion.extra_goals) -> query
 
 val total_type_checking_time : query -> float
+val total_det_checking_time : query -> float
 
 val optimize_query : query -> executable
 
@@ -56,8 +58,9 @@ val pp_goal : (pp_ctx:pp_ctx -> depth:int -> Format.formatter -> term -> unit) -
 val elpi_language : Compiler_data.Scope.language
 val elpi : Compiler_data.QuotationHooks.quotation
 
-val uvk : uvar_body StrMap.t State.component
-val pp : Format.formatter -> uvar_body -> unit
+val uvk : uvar StrMap.t State.component
+val pp_uvar_body : Format.formatter -> uvar -> unit
+val pp_uvar_body_raw : Format.formatter -> uvar -> unit
 
 val compile_term_to_raw_term :
   ?check:bool -> State.t -> program ->
@@ -67,4 +70,18 @@ val runtime_hack_term_to_raw_term :
   State.t -> program ->
   ?ctx:constant Compiler_data.Scope.Map.t ->
   depth:int -> Compiler_data.ScopedTerm.t -> term
-val global_name_to_constant : State.t -> string -> constant
\ No newline at end of file
+val global_name_to_constant : State.t -> string -> constant
+
+module IntervalTree : sig
+  type 'a t
+  val find : Ast.Loc.t -> 'a t -> (Ast.Loc.t * 'a) list
+  val pp : (Format.formatter -> 'a -> unit) -> Format.formatter -> 'a t -> unit
+end
+
+type type_
+val pp_type_ : Format.formatter -> type_ -> unit
+
+type info = { defined : Ast.Loc.t option; type_ : type_ option }
+val pp_info : Format.formatter -> info -> unit
+
+val hover : checked_compilation_unit -> info IntervalTree.t
