@@ -11,6 +11,61 @@ module Fmt = Format
 module F = Ast.Func
 open Util
 
+(** 
+  Used to index the parameters of a predicate P
+  - [MapOn N] -> N-th argument at depth 1 (head symbol only)
+  - [Hash L]  -> L is the list of depths given by the urer for the parameters of
+                 P. Indexing is done by hashing all the parameters with a non
+                 zero depth and comparing it with the hashing of the parameters
+                 of the query
+  - [DiscriminationTree L] ->
+            we use the same logic of Hash, except we use DiscriminationTree to discriminate
+            clauses
+*)
+type indexing =
+  | MapOn of int
+  | Hash of int list
+  | DiscriminationTree of int list
+[@@deriving show, ord]
+
+(* 
+Used in the lightweight discrimination tree to
+statically verify whether two rules overlap.
+
+- has_cut: 
+    Instead of storing the full body of the clause in the 
+    lightweight discrimination tree, we simply record 
+    whether the rule contains a cut.
+
+- arg_nb: 
+    In the discrimination tree, variadic arguments are not included in the path.
+    However, the number of arguments is used as a discriminator between two rules.
+    arg_nb tells the number of arguments in the clause when it was added to the trie.
+*)
+type overlap_clause = { 
+  overlap_loc : Loc.t option; has_cut : bool; mutable timestamp : int list;
+  arg_nb : int
+}
+[@@deriving show]
+
+type overlap =
+  | Allowed
+  | Forbidden of overlap_clause Discrimination_tree.t
+[@@deriving show]
+
+let mk_Forbidden indexing =
+  Forbidden (
+    match indexing with
+    | MapOn i -> Discrimination_tree.empty_dt (List.init (i+1) (fun j -> if j = i then 1 else 0))
+    | Hash l -> Discrimination_tree.empty_dt l
+    | DiscriminationTree l -> Discrimination_tree.empty_dt l
+  )
+
+type pred_info = { indexing:indexing; mode:Mode.hos; overlap: overlap; has_local_without_cut: Loc.t option }
+[@@deriving show]
+
+let same_indexing { indexing = i1 } { indexing = i2 } = compare_indexing i1 i2 == 0
+
 (******************************************************************************
   Terms: data type definition and printing
  ******************************************************************************)
@@ -55,44 +110,103 @@ let equal_stuck_goal_kind _ x y = x == y
 type 'unification_def stuck_goal_kind +=
   | Unification of 'unification_def
 
-type arg_mode = Util.arg_mode = Input | Output [@@deriving show, ord]
-
-type mode_aux = Util.mode_aux =
-  | Fo of arg_mode
-  | Ho of arg_mode * mode
-and mode = mode_aux list
-[@@ deriving show, ord]
-
 type ttype =
   | TConst of constant
   | TApp of constant * ttype * ttype list
-  | TPred of bool * ((arg_mode * ttype) list) (* The bool is for functionality *)
+  | TPred of bool * ((Mode.t * ttype) list) (* The bool is for functionality *)
   | TArr of ttype * ttype
   | TCData of CData.t
   | TLam of ttype (* this is for parametrized typeabbrevs *)
   [@@ deriving show, ord]
 
+type builtin_predicate =
+  | Cut | And | Impl | ImplBang | RImpl | Pi | Sigma | Eq | Match | Findall | Delay | Host of constant [@@deriving ord, show]
+let builtin_predicates = [Cut;And;Impl;ImplBang;RImpl;Pi;Sigma;Eq;Match;Findall;Delay]
+
+let builtin_predicate_max = List.length builtin_predicates
+let func_of_builtin_predicate f = function
+  | Cut     -> F.cutf
+  | And     -> F.andf
+  | Impl    -> F.implf
+  | ImplBang -> F.implbangf
+  | RImpl   -> F.rimplf
+  | Pi      -> F.pif
+  | Sigma   -> F.sigmaf
+  | Eq      -> F.eqf
+  | Match   -> F.pmf
+  | Findall -> F.from_string "findall_solutions"
+  | Delay   -> F.from_string "declare_constraint"
+  | Host c  -> f c
+
+let show_builtin_predicate ?table f = function
+  | Host c -> f ?table c
+  | x -> F.show (func_of_builtin_predicate (fun _ -> assert false) x)
+
+let const_of_builtin_predicate = function
+  | Cut     -> -1
+  | And     -> -2
+  | Impl    -> -3
+  | RImpl   -> -4
+  | Pi      -> -5
+  | Sigma   -> -6
+  | Eq      -> -7
+  | Match   -> -8
+  | Findall -> -9
+  | Delay   -> -10
+  | ImplBang -> -11
+  | Host c -> c
+  
+let is_builtin_predicate c = - builtin_predicate_max <= c && c <= -1
+
+let builtin_predicate_of_const = function
+  | -1 -> Cut   
+  | -2 -> And   
+  | -3 -> Impl  
+  | -4 -> RImpl 
+  | -5 -> Pi    
+  | -6 -> Sigma 
+  | -7 -> Eq    
+  | -8 -> Match
+  | -9 -> Findall
+  | -10 -> Delay
+  | -11 -> ImplBang
+  | _ -> assert false
+  
+let () = assert(List.for_all (fun p -> is_builtin_predicate @@ const_of_builtin_predicate p) builtin_predicates)
+let () = assert(List.for_all (fun p -> p = builtin_predicate_of_const @@ const_of_builtin_predicate p) builtin_predicates)
+
+
+let map_builtin_predicate f = function
+  | Host x -> Host (f x)
+  | x -> x
+
 type term =
   (* Pure terms *)
   | Const of constant
   | Lam of term
   | App of constant * term * term list
+
   (* Optimizations *)
   | Cons of term * term
   | Nil
   | Discard
+
   (* FFI *)
-  | Builtin of constant * term list
+  | Builtin of builtin_predicate * term list
   | CData of CData.t
+
   (* Heap terms: unif variables in the query *)
-  | UVar of uvar_body * (*depth:*)int * (*argsno:*)int
-  | AppUVar of uvar_body * (*depth:*)int * term list
-  (* Clause terms: unif variables used in clauses *)
-  | Arg of (*id:*)int * (*argsno:*)int
-  | AppArg of (*id*)int * term list
-and uvar_body = {
+  | UVar    of uvar * int (* number of arguments *)
+  | AppUVar of uvar * term list (* arguments *)
+
+  (* Stack terms: unif variables in the rules *)
+  | Arg    of int * int (* number of arguments *)
+  | AppArg of int * term list (* arguments *)
+
+and uvar = {
+  vardepth : int; (* depth at creation time *)
   mutable contents : term [@printer (pp_spaghetti_any ~id:id_term pp_oref)];
-  mutable uid_private : int; (* unique name, the sign is flipped when blocks a constraint *)
+  mutable uid_private : int; (* negative if it blocks a constraint *)
 }
 [@@deriving show, ord]
 
@@ -111,15 +225,12 @@ type clause = {
     args : term list;
     hyps : term list;
     vars : int;
-    mode : mode;        (* CACHE to avoid allocation in get_clauses *)
+    mode : Mode.hos;        (* CACHE to avoid allocation in get_clauses *)
     loc : Loc.t option; (* debug *)
     mutable timestamp : int list; (* for grafting *)
 }
 [@@deriving show, ord]
 
-let get_arg_mode = function Fo a -> a | Ho (a,_) -> a 
-let to_mode = function true -> Fo Input | false -> Fo Output
-
 type grafting_time = int list
 [@@deriving show, ord]
 let compare_constant = Util.compare_constant
@@ -130,7 +241,7 @@ type stuck_goal = {
   mutable blockers : blockers;
   kind : unification_def stuck_goal_kind;
 }
-and blockers = uvar_body list
+and blockers = uvar list
 and unification_def = {
   adepth : int;
   env : term array;
@@ -156,19 +267,19 @@ and first_lvl_idx = {
 }
 and second_lvl_idx =
 | TwoLevelIndex of {
-    mode : mode;
+    mode : Mode.hos;
     argno : int;
     all_clauses : clause_list;        (* when the query is flexible *)
-    flex_arg_clauses : clause_list;   (* when the query is rigid but arg_id ha nothing *)
+    flex_arg_clauses : clause_list;   (* when the query is rigid but arg_id has nothing *)
     arg_idx : clause_list Ptmap.t;    (* when the query is rigid (includes in each binding flex_arg_clauses) *)
   }
 | BitHash of {
-    mode : mode;
+    mode : Mode.hos;
     args : int list;
     args_idx : clause_list Ptmap.t; (* clause, insertion time *)
   }
 | IndexWithDiscriminationTree of {
-    mode : mode;
+    mode : Mode.hos;
     arg_depths : int list;   (* the list of args on which the trie is built *)
     args_idx : clause Discrimination_tree.t; 
 }
@@ -183,26 +294,10 @@ type hyps = clause_src list
 
 type suspended_goal = {
   context : hyps;
-  goal : int * term
+  goal : int * term;
+  blockers : blockers;
 }
 
-(** 
-  Used to index the parameters of a predicate P
-  - [MapOn N] -> N-th argument at depth 1 (head symbol only)
-  - [Hash L]  -> L is the list of depths given by the urer for the parameters of
-                 P. Indexing is done by hashing all the parameters with a non
-                 zero depth and comparing it with the hashing of the parameters
-                 of the query
-  - [DiscriminationTree L] ->
-            we use the same logic of Hash, except we use DiscriminationTree to discriminate
-            clauses
-*)
-type indexing =
-  | MapOn of int
-  | Hash of int list
-  | DiscriminationTree of int list
-[@@deriving show]
-
 let mkLam x = Lam x [@@inline]
 let mkApp c x xs = App(c,x,xs) [@@inline]
 let mkCons hd tl = Cons(hd,tl) [@@inline]
@@ -210,8 +305,8 @@ let mkNil = Nil
 let mkDiscard = Discard
 let mkBuiltin c args = Builtin(c,args) [@@inline]
 let mkCData c = CData c [@@inline]
-let mkUVar r d ano = UVar(r,d,ano) [@@inline]
-let mkAppUVar r d args = AppUVar(r,d,args) [@@inline]
+let mkUVar r ano = UVar(r,ano) [@@inline]
+let mkAppUVar r args = AppUVar(r,args) [@@inline]
 let mkArg i ano = Arg(i,ano) [@@inline]
 let mkAppArg i args = AppArg(i,args) [@@inline]
 
@@ -248,7 +343,7 @@ let destConst = function Const x -> x | _ -> assert false
    After the constraint store, since assigning may wake up some constraints *)
 let oref =
   let uid = ref 0 in
-  fun x -> incr uid; assert(!uid > 0); { contents = x; uid_private = !uid }
+  fun ~depth:vardepth x -> incr uid; assert(!uid > 0); { vardepth; contents = x; uid_private = !uid }
 let (!!) { contents = x } = x
 
 (* Arg/AppArg point to environments, here the empty one *)
@@ -325,7 +420,6 @@ end = struct
 
   type t = { data : Obj.t StrMap.t; stage : stage; extensions : descriptor }
   let dummy : t = { data = StrMap.empty; stage = Dummy; extensions = ref StrMap.empty }
-  let descriptor { extensions = x } = x
  
   let new_descriptor () : descriptor = ref StrMap.empty
   let merge_descriptors m1 m2 =
@@ -421,6 +515,181 @@ end
 
 let elpi_state_descriptor = State.new_descriptor ()
 
+type core_symbol = As | Uv | ECons | ENil [@@deriving enum, ord, show]
+let func_of_core_symbol = function
+  | As    -> F.asf
+  | Uv    -> F.from_string "uvar"
+  | ECons  -> F.consf
+  | ENil   -> F.nilf
+let is_core_symbol f =
+  let rec aux i =
+    if i < max_core_symbol then
+      F.equal f (func_of_core_symbol (Option.get (core_symbol_of_enum i))) || aux (i+1)
+    else
+      false
+  in
+      aux 0
+
+(* Globally unique identifier for symbols with a quotient *)
+module Symbol : sig 
+  type symbol [@@deriving show]
+  module UF : Union_find.S with type key = symbol
+  type 'a merge = (symbol -> 'a -> 'a -> 'a)
+  module RawMap : Map.S with type key = symbol
+  module QMap : sig
+    type 'a t [@@deriving show]
+    val empty : 'a t
+    val add : symbol -> 'a -> 'a t -> 'a t
+    val find : symbol -> 'a t -> 'a
+    val find_opt : symbol -> 'a t -> 'a option
+    val union : 'a merge -> 'a t -> 'a t -> 'a t
+    val give_uf : 'a t -> UF.t
+    val unify : 'a merge -> symbol -> symbol -> 'a t -> 'a t
+    val mapi : (symbol -> symbol) -> ('a -> 'a) -> 'a t -> 'a t
+    val map : ('a -> 'b) -> 'a t -> 'b t
+    val fold : (symbol -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
+    val iter : (symbol -> 'a -> unit) -> 'a t -> unit
+    val mem : symbol -> 'a t -> bool
+    val bindings : 'a t -> (symbol * 'a) list
+    val get_uf : 'a t -> UF.t
+  end 
+
+  type t = symbol [@@deriving show]
+  type provenance = Elpi_parser.Ast.Structured.provenance [@@deriving show,ord]
+
+  val equal : uf:UF.t -> t -> t -> bool
+  val compare : uf:UF.t -> t -> t -> int
+
+  val make : provenance -> F.t -> t
+  val make_builtin : ?variant:int -> F.t -> t
+  val make_variant_builtin : F.t -> t * int (* variant *)
+
+  val get_loc : t -> Loc.t
+  val get_provenance : t -> provenance
+  val get_str : t -> string
+  val get_func : t -> F.t
+  val is_builtin : t -> F.t -> bool
+
+  val undup : uf:UF.t -> t list -> t list
+
+  val pretty : t -> string
+
+  (* val map_func : (F.t -> F.t) -> t -> t *)
+  
+end = struct
+  type provenance = Elpi_parser.Ast.Structured.provenance [@@deriving show,ord]
+  type symbol = provenance * F.t [@@deriving show, ord]
+  let compare_symbol (p1,f1) (p2,f2) = let x = F.compare f1 f2 in if x = 0 then compare_provenance p1 p2 else x
+  type 'a merge = (symbol -> 'a -> 'a -> 'a)
+  module O = struct type t = symbol [@@deriving show,ord] end
+  module RawMap = Map.Make(O)
+  module UF = Elpi_util.Union_find.Make(O)
+  type t = symbol [@@deriving show]
+
+  open Elpi_parser.Ast.Structured
+
+  let pretty (prov,f) = F.show f ^ match prov with
+    | Core -> " (core symbol)"
+    | File _ -> ""
+    | Builtin { variant } -> if variant <> 0 then " (variant "^string_of_int variant ^")" else ""
+
+  module QMap = struct
+    
+    type 'a t = UF.t * 'a RawMap.t [@@deriving show]
+    let empty = UF.empty, RawMap.empty
+
+    let get_uf (u,_) = u
+
+    let unify f s1 s2 (uf,m) = 
+      let x,uf =
+        match fst s1, fst s2 with
+        | Builtin _, Builtin _ -> anomaly "Builtins cannot be declared twice"
+        | Core, Core -> anomaly "Core symbols cannot be declared twice"
+        (* we use the (possibly) pre-allocated builtin as the canonical *)
+        | File _, Builtin _ -> UF.union uf s1 ~canon:s2
+        | File _, Core -> UF.union uf s1 ~canon:s2
+        | _ -> UF.union uf ~canon:s1 s2 in
+      match x with
+      | None -> uf, m
+      | Some x -> 
+        let canonic_x = UF.find uf x in
+        match RawMap.find_opt canonic_x m, RawMap.find_opt x m with
+        | None, None -> uf, m
+        | Some canonic_value, None -> uf, m
+        | None, Some value -> uf, RawMap.add canonic_x value m |> RawMap.remove x
+        | Some canonic_value, Some value -> 
+          uf, RawMap.add canonic_x (f canonic_x value canonic_value) m 
+
+    let add s a (uf,m) =
+      let s' = UF.find uf s in
+      uf, RawMap.add s' a m
+
+    let find s (uf,m) =
+      let s' = UF.find uf s in
+      try RawMap.find s' m
+      with Not_found -> anomaly ("Missing entry from QMap: " ^ show_symbol s)
+
+    let find_opt s (uf,m) = RawMap.find_opt (UF.find uf s) m
+
+    let union f (uf1,m1) (uf2,m2) =
+      let uf = UF.merge uf1 uf2 in
+      uf,RawMap.union (fun s a b -> Some (f (UF.find uf s) a b)) m1 m2
+
+    let map mv (uf, m) =
+      uf, RawMap.map mv m
+
+    let mapi mk mv (uf,m) =
+      let uf = UF.mapi mk uf in
+      let m = RawMap.fold (fun k v acc -> RawMap.add (mk k) (mv v) acc) RawMap.empty m in
+      uf,m
+
+    let give_uf (a,_) = a
+    let fold f (_,m) a = RawMap.fold f m a
+
+    let mem s (uf,m) =
+      let s' = UF.find uf s in
+      RawMap.mem s' m
+
+    let iter f (_,m) = RawMap.iter f m
+
+    let bindings (_,m) = RawMap.bindings m
+
+  end
+
+  let equal ~uf x y = compare (UF.find uf x) (UF.find uf y) = 0
+  let compare ~uf x y = compare (UF.find uf x) (UF.find uf y)
+
+  let rec undup ~uf = function
+  | [] -> []
+  | x :: xs -> let x = UF.find uf x in if List.exists (fun y -> compare ~uf x y = 0) xs then undup ~uf xs else x :: undup ~uf xs
+
+  let is_builtin (p,f) s =
+    F.equal f s && match p with Builtin { variant } -> variant = 0 | _ -> false
+
+  let get_provenance (l,_) = l
+  let get_loc (l,f) =
+    match l with
+    | File l -> l
+    | Core -> Loc.initial ("("^__FILE__^":"^F.show f^")")
+    | Builtin { variant } -> Loc.initial ("(ocaml:"^F.show f^":"^string_of_int variant^")")
+  let get_str (_,f) = F.show f
+  let get_func (_,f) = f
+  
+  let make prov name = prov, name
+  let make_builtin ?(variant=0) name = Builtin { variant }, name
+  let make_variant_builtin =
+    let state = ref F.Map.empty in
+    let incr name =
+      let n = try F.Map.find name !state with Not_found -> 0 in
+      let n = n + 1 in
+      state := F.Map.add name n !state;
+      n in
+    fun name ->
+      let variant = incr name in
+      (Builtin { variant }, name), variant
+end
+
+
 (* This module contains the symbols reserved by Elpi and the ones
    declared by the API client via declare_global_symbol statically
    (eg the API must be called in a OCaml toplevel value).
@@ -434,8 +703,8 @@ module Global_symbols : sig
   (* Table used at link time *)
   type t = {
     (* Ast (functor name) -> negative int n (constant) * hashconsed (Const n) *)
-    mutable s2ct : (constant * term) Ast.Func.Map.t;
-    mutable c2s : string Constants.Map.t;
+    mutable s2ct : (constant * term) Symbol.RawMap.t;
+    mutable c2s : Symbol.symbol Constants.Map.t;
     (* negative *)
     mutable last_global : int;
 
@@ -445,113 +714,117 @@ module Global_symbols : sig
   val table : t
 
   (* Static initialization, eg link time *)
-  val declare_global_symbol : string -> constant
+  val declare_global_symbol : ?variant:int -> string -> constant
+  (* Used by auto-generated code, eg ADT *)
+  val declare_overloaded_global_symbol : string -> constant * int
   val lock : unit -> unit
 
-  val cutc     : constant
-  val andc     : constant
-  val orc      : constant
-  val implc    : constant
-  val rimplc   : constant
-  val pic      : constant
-  val sigmac   : constant
-  val eqc      : constant
-  val rulec    : constant
-  val consc    : constant
-  val nilc     : constant
-  val entailsc : constant
-  val nablac   : constant
-  val asc      : constant
-  val arrowc   : constant
-  val uvarc    : constant
-  val propc    : constant
+  open Symbol
 
-  val ctypec   : constant
-  val variadic : constant
+  (* builtin predicates *)
+  val cut : symbol
+  val and_ : symbol
+  val impl : symbol
+  val rImpl : symbol
+  val pi : symbol
+  val sigma : symbol
+  val eq : symbol
+  val match_ : symbol
+  val findall : symbol
+  val delay : symbol
 
-  val spillc   : constant
-  val truec    : constant
+  (* core symbols *)
+  val as_      : symbol
+  val uvar    : symbol
+  val nil    : symbol
+  val cons    : symbol
 
-  val declare_constraintc : constant
-  val print_constraintsc  : constant
-  val findall_solutionsc  : constant
+  (* internal *)
+  val uvarc : constant (* needed by runtime/unif *)
+  val asc : constant (* needed by runtime/unif *)
+  val orc      : constant (* needed by coq-elpi *)
+  val nilc     : constant (* needed by indexing *)
+  val consc     : constant (* needed by indexing *)
 
 end = struct
 
 type t = {
-  mutable s2ct : (constant * term) Ast.Func.Map.t;
-  mutable c2s : string Constants.Map.t;
+  mutable s2ct : (constant * term) Symbol.RawMap.t;
+  mutable c2s : Symbol.t Constants.Map.t;
   mutable last_global : int;
   mutable locked : bool;
 }
 [@@deriving show]
 
 let table = {
-  last_global = 0;
-  s2ct = Ast.Func.Map.empty;
+  last_global = - builtin_predicate_max;
+  s2ct = Symbol.RawMap.empty;
   c2s = Constants.Map.empty;
   locked = false;
 }
 
-let declare_global_symbol str =
-  let x = Ast.Func.from_string str in
-  try fst @@ Ast.Func.Map.find x table.s2ct
+let () = builtin_predicates |> List.iter (fun p ->
+  let c = const_of_builtin_predicate p in
+  let s = Symbol.make_builtin (func_of_builtin_predicate (fun _ -> assert false) p) in
+  let t = Const c in
+  table.c2s <- Constants.Map.add c s table.c2s;
+  table.s2ct <- Symbol.RawMap.add s (c,t) table.s2ct
+)
+
+let declare_global_symbol symb =
+  try fst @@ Symbol.RawMap.find symb table.s2ct
   with Not_found ->
     if table.locked then
       Util.anomaly "declare_global_symbol called after initialization";
     table.last_global <- table.last_global - 1;
     let n = table.last_global in
     let t = Const n in
-    table.s2ct <- Ast.Func.Map.add x (n,t) table.s2ct;
-    table.c2s <- Constants.Map.add n str table.c2s;
+    table.s2ct <- Symbol.RawMap.add symb (n,t) table.s2ct;
+    table.c2s <- Constants.Map.add n symb table.c2s;
     n
 
-let declare_global_symbol_for_builtin str =
-  let x = Ast.Func.from_string str in
-  if table.locked then
-    Util.anomaly "declare_global_symbol_for_builtin called after initialization";
-  try fst @@ Ast.Func.Map.find x table.s2ct
-  with Not_found ->
-    table.last_global <- table.last_global - 1;
-    let n = table.last_global in
-    let t = Builtin(n,[]) in
-    table.s2ct <- Ast.Func.Map.add x (n,t) table.s2ct;
-    table.c2s <- Constants.Map.add n str table.c2s;
-    n
+
+let declare_core_symbol x =
+  let symb = Symbol.(make Core (func_of_core_symbol x)) in
+  declare_global_symbol symb, symb
+let uvarc, uvar = declare_core_symbol Uv
+let asc, as_ = declare_core_symbol As
+let nilc, nil = declare_core_symbol ENil
+let consc, cons = declare_core_symbol ECons
+
+let declare_overloaded_global_symbol str =
+  let symb, variant = Symbol.make_variant_builtin (Ast.Func.from_string str) in
+  declare_global_symbol symb, variant
+
+let declare_global_symbol ?variant str =
+  let symb = Symbol.make_builtin ?variant (Ast.Func.from_string str) in
+  declare_global_symbol symb
 
 let lock () = table.locked <- true
 
-let andc                = declare_global_symbol F.(show andf)
-let arrowc              = declare_global_symbol F.(show arrowf)
-let asc                 = declare_global_symbol "as"
-let consc               = declare_global_symbol F.(show consf)
-let entailsc            = declare_global_symbol "?-"
-let eqc                 = declare_global_symbol F.(show eqf)
-let uvarc               = declare_global_symbol "uvar"
-let implc               = declare_global_symbol F.(show implf)
-let nablac              = declare_global_symbol "nabla"
-let nilc                = declare_global_symbol F.(show nilf)
 let orc                 = declare_global_symbol F.(show orf)
-let pic                 = declare_global_symbol F.(show pif)
-let rimplc              = declare_global_symbol F.(show rimplf)
-let rulec               = declare_global_symbol "rule"
-let sigmac              = declare_global_symbol F.(show sigmaf)
-let spillc              = declare_global_symbol F.(show spillf)
-let truec               = declare_global_symbol F.(show truef)
-let ctypec              = declare_global_symbol F.(show ctypef)
-let propc               = declare_global_symbol "prop"
-let variadic            = declare_global_symbol "variadic"
 
-let declare_constraintc = declare_global_symbol "declare_constraint"
-let cutc                = declare_global_symbol_for_builtin F.(show cutf)
-let print_constraintsc  = declare_global_symbol_for_builtin "print_constraints"
-let findall_solutionsc  = declare_global_symbol_for_builtin "findall_solutions"
+let publish_builtin b = Constants.Map.find (const_of_builtin_predicate b) table.c2s 
+
+let cut = publish_builtin Cut
+let and_ = publish_builtin And
+let impl = publish_builtin Impl
+let rImpl = publish_builtin RImpl
+let pi = publish_builtin Pi
+let sigma = publish_builtin Sigma
+let eq = publish_builtin Eq
+let match_ = publish_builtin Match
+let findall = publish_builtin Findall
+let delay = publish_builtin Delay
+
 
 end
 
 (* This term is hashconsed here *)
-let dummy = App (Global_symbols.cutc,Const Global_symbols.cutc,[])
-let dummy_uvar_body = { contents = dummy; uid_private = 0 }
+let dummy = App (-1,Const (-1),[])
+
+(* This is the one uvar used for _ in CHR rules blockers *)
+let dummy_uvar_body = { vardepth = 0; contents = dummy; uid_private = 0 }
 
 module CHR : sig
 
@@ -674,6 +947,7 @@ type clause_w_info = {
 
 exception No_clause
 exception No_more_steps
+exception Flex_head
 
 
 module Conversion = struct
@@ -743,7 +1017,7 @@ let rec show_ty_ast ?prec = function
       with_par prec AppArg t
 
 let term_of_extra_goal = function
-  | Unify(a,b) -> Builtin(Global_symbols.eqc,[a;b])
+  | Unify(a,b) -> Builtin(Eq,[a;b])
   | RawGoal x -> x
   | x ->
       Util.anomaly (Printf.sprintf "Unprocessed extra_goal: %s.\nOnly %s and %s can be left unprocessed,\nplease call API.RawData.set_extra_goals_postprocessing.\n"
@@ -787,13 +1061,14 @@ module ContextualConversion = struct
     embed = (fun ~depth s t -> embed ~depth () () s t);
     readback = (fun ~depth s t -> readback ~depth () () s t);
   }
-
   let (!>) { Conversion.ty; pp_doc; pp; embed; readback; } = {
     ty; pp; pp_doc;
     embed = (fun ~depth _ _ s t -> embed ~depth s t);
     readback = (fun ~depth _ _ s t -> readback ~depth s t);
   }
 
+  let (!<<) f x = (!<) (f ((!>) x))
+
   let (!>>) (f : 'a Conversion.t -> 'b Conversion.t) cc =
   let mk h c { ty; pp_doc; pp; embed; readback; } = {
     Conversion.ty; pp; pp_doc;
@@ -968,13 +1243,20 @@ type ('t,'h,'c) constructor =
       ('match_stateful_t,'match_t,'t) match_t
     -> ('t,'h,'c) constructor
 
-type ('t,'h,'c) declaration = {
+type ('t,'h,'c) base_declaration = {
   ty : Conversion.ty_ast;
   doc : doc;
   pp : Format.formatter -> 't -> unit;
   constructors : ('t,'h,'c) constructor list;
 }
 
+type ('t,'h,'c) declaration =
+| Decl : ('t,'h,'c) base_declaration -> ('t,'h,'c) declaration
+| Param : ('t Conversion.t -> ('t1,'h,'c) declaration) -> ('t1,'h,'c) declaration
+| ParamC : (('t,'h,'c) ContextualConversion.t -> ('t1,'h,'c) declaration) -> ('t1,'h,'c) declaration
+
+type allocation = (constant * int) StrMap.t
+
 type ('b,'m,'t,'h,'c) compiled_constructor_arguments =
   | XN : (State.t -> State.t * 't,State.t -> State.t * term * Conversion.extra_goals, 't,'h,'c) compiled_constructor_arguments
   | XA : ('a,'h,'c) ContextualConversion.t * ('b,'m,'t,'h,'c) compiled_constructor_arguments -> ('a -> 'b, 'a -> 'm, 't,'h,'c) compiled_constructor_arguments
@@ -1138,21 +1420,37 @@ let rec tyargs_of_args : type a b c d e. string -> (a,b,c,d,e) compiled_construc
   | XN -> [false,self,""]
   | XA ({ ty },rest) -> (false,Conversion.show_ty_ast ty,"") :: tyargs_of_args self rest
 
-let compile_constructors ty self self_name l =
+let do_allocate_constructors ty l =
   let names =
     List.fold_right (fun (K(name,_,_,_,_)) -> StrSet.add name) l StrSet.empty in
   if StrSet.cardinal names <> List.length l then
     anomaly ("Duplicate constructors name in ADT: " ^ Conversion.show_ty_ast ty);
-  List.fold_left (fun (acc,sacc) (K(name,_,a,b,m)) ->
-    let c = Global_symbols.declare_global_symbol name in
-    let args = compile_arguments a self in
-    Constants.Map.add c (XK(args,compile_builder a b,compile_matcher a m)) acc,
-    StrMap.add name (tyargs_of_args self_name args) sacc)
-      (Constants.Map.empty,StrMap.empty) l
+  List.fold_left (fun vacc (K(name,_,a,b,m)) ->
+    if name = "uvar" then
+      vacc
+    else
+      let c_variant = Global_symbols.declare_overloaded_global_symbol name in
+      StrMap.add name c_variant vacc)
+    StrMap.empty l
 
-let document_constructor fmt name doc argsdoc =
-  Fmt.fprintf fmt "@[<hov2>type %s@[<hov>%a.%s@]@]@\n"
-    name pp_ty_args argsdoc (if doc = "" then "" else " % " ^ doc)
+let compile_constructors allocated ty self self_name l =
+  List.fold_left (fun (acc, sacc) (K(name,_,a,b,m)) ->
+    if name = "uvar" then
+      let args = compile_arguments a self in
+      let acc = Constants.Map.add Global_symbols.uvarc (XK(args,compile_builder a b,compile_matcher a m)) acc in
+      (acc, sacc)
+    else
+      let c =
+        try StrMap.find name allocated |> fst
+        with Not_found -> anomaly "constructor arguments should be preallocated" in
+      let args = compile_arguments a self in
+             Constants.Map.add c (XK(args,compile_builder a b,compile_matcher a m)) acc,
+      StrMap.add name (tyargs_of_args self_name args) sacc)
+        (Constants.Map.empty,StrMap.empty) l
+
+let document_constructor fmt name variant doc argsdoc =
+  Fmt.fprintf fmt "@[<hov2>external symbol %s :@[<hov>%a@] = \"%d\". %s@]@\n"
+    name pp_ty_args argsdoc variant (if doc = "" then "" else " % " ^ doc)
 
 let document_kind fmt = function
   | Conversion.TyApp(s,_,l) ->
@@ -1162,16 +1460,46 @@ let document_kind fmt = function
         s (String.concat " -> " (Array.to_list l))
   | Conversion.TyName s -> Fmt.fprintf fmt "@[<hov 2>kind %s type.@]@\n" s
 
-let document_adt doc ty ks cks fmt () =
+let document_adt doc ty ks cks vks fmt () =
   if doc <> "" then
     begin pp_comment fmt ("% " ^ doc); Fmt.fprintf fmt "@\n" end;
   document_kind fmt ty;
   List.iter (fun (K(name,doc,_,_,_)) ->
     if name <> "uvar" then
       let argsdoc = StrMap.find name cks in
-      document_constructor fmt name doc argsdoc) ks
+      document_constructor fmt name (StrMap.find name vks |> snd) doc argsdoc) ks
 
-let adt ~mkinterval ~look ~mkConst ~alloc ~mkUnifVar { ty; constructors; doc; pp } =
+let rec allocate_constructors: type t h c. mkinterval:(int -> int -> int -> term list) -> look:(depth:int -> term -> term) -> mkConst:(int -> term) -> alloc:(?name:doc -> State.t -> State.t * 'a) -> mkUnifVar:
+    ('a -> args:term list -> State.t -> term) ->
+      (t,h,c) declaration -> allocation =
+      fun ~mkinterval ~look ~mkConst ~alloc ~mkUnifVar ->
+function
+| Decl { ty; constructors; doc; pp } -> do_allocate_constructors ty constructors
+| Param f ->
+  let a = {
+    ContextualConversion.ty = Conversion.TyName "A";
+    pp = (fun fmt _ -> ());
+    pp_doc = (fun fmt () -> ());
+    readback = (fun ~depth hyps constraints state term -> assert false);
+    embed = (fun ~depth hyps constraints state term -> assert false);
+  } |> ContextualConversion.(!<) in
+  allocate_constructors ~mkinterval ~look ~mkConst ~alloc ~mkUnifVar (f a)
+| ParamC f ->
+  let a = {
+    ContextualConversion.ty = Conversion.TyName "A";
+    pp = (fun fmt _ -> ());
+    pp_doc = (fun fmt () -> ());
+    readback = (fun ~depth hyps constraints state term -> assert false);
+    embed = (fun ~depth hyps constraints state term -> assert false);
+  } in
+  allocate_constructors ~mkinterval ~look ~mkConst ~alloc ~mkUnifVar (f a)
+
+
+  let declare_allocated: type t h c. mkinterval:(int -> int -> int -> term list) -> look:(depth:int -> term -> term) -> mkConst:(int -> term) -> alloc:(?name:doc -> State.t -> State.t * 'a) -> mkUnifVar:
+  ('a -> args:term list -> State.t -> term) ->
+    allocation -> (t,h,c) declaration -> (t,h,c) ContextualConversion.t =
+    fun ~mkinterval ~look ~mkConst ~alloc ~mkUnifVar allocated ->
+function Decl { ty; constructors; doc; pp } ->
   let readback_ref = ref (fun ~depth _ _ _ _ -> assert false) in
   let embed_ref = ref (fun ~depth _ _ _ _ -> assert false) in
   let sconstructors_ref = ref StrMap.empty in
@@ -1179,17 +1507,18 @@ let adt ~mkinterval ~look ~mkConst ~alloc ~mkUnifVar { ty; constructors; doc; pp
     ContextualConversion.ty;
     pp;
     pp_doc = (fun fmt () ->
-      document_adt doc ty constructors !sconstructors_ref fmt ());
+      document_adt doc ty constructors !sconstructors_ref allocated fmt ());
     readback = (fun ~depth hyps constraints state term ->
       !readback_ref ~depth hyps constraints state term);
     embed = (fun ~depth hyps constraints state term ->
       !embed_ref ~depth hyps constraints state term);
   } in
-  let cconstructors, sconstructors = compile_constructors ty self (Conversion.show_ty_ast ty) constructors in
+  let cconstructors, sconstructors = compile_constructors allocated ty self (Conversion.show_ty_ast ty) constructors in
   sconstructors_ref := sconstructors;
   readback_ref := readback ~mkinterval ~look ~alloc ~mkUnifVar ty cconstructors;
   embed_ref := embed ~mkConst ty pp cconstructors;
   self
+| _ -> anomaly "declare_allocated can only be called on Decl"
 
 end
 
@@ -1201,53 +1530,107 @@ type declaration =
   | LPCode of string
 
 (* doc *)
+let parens ?(sep = " ") str = if Re.Str.(string_match (regexp (".*" ^ sep ^ ".*")) str 0) then "("^str^")" else str
+let parens_arr = parens ~sep:("->")
+
 let ws_to_max fmt max n =
   if n < max then Format.fprintf fmt "%s" (String.make (max - n) ' ')
   else ()
-let pp_tab_arg i max sep fmt (dir,ty,doc) =
-  let dir = if dir then "i" else "o" in
+
+let pp_tab_arg i max pre sep fmt (_,ty,doc) =
   if i = 0 then Fmt.pp_set_tab fmt () else ();
-  Fmt.fprintf fmt "%s:%s%s" dir ty sep;
+  Fmt.fprintf fmt "%s%s%s" pre (parens_arr ty) sep;
   if i = 0 then (ws_to_max fmt max (String.length ty); Fmt.pp_set_tab fmt ()) else Fmt.pp_print_tab fmt ();
   if doc <> "" then begin Fmt.fprintf fmt " %% %s" doc end;
   Fmt.pp_print_tab fmt ()
 ;;
 
 let pp_tab_args fmt l =
-  let n = List.length l - 1 in
   let max = List.fold_left (fun m (_,s,_) -> max (String.length s) m) 0 l in
   Fmt.pp_open_tbox fmt ();
-  if l = [] then Fmt.fprintf fmt ".";
-  List.iteri (fun i x ->
-    let sep = if i = n then "." else "," in
-    pp_tab_arg i max sep fmt x) l;
+
+  let rec aux m_of_last i = function
+    | [] -> ()
+    | x :: xs ->
+        let (m_of_x,_,_) = x in
+        let pre = if m_of_last <> m_of_x then "-> " else "" in
+        let sep = if xs = [] then "." else
+          match xs with
+          | (b,_,_) :: _ when b = m_of_x -> ", "
+          | _ -> "" in
+        pp_tab_arg i max pre sep fmt x;
+        aux m_of_x (i+1) xs in
+
+  if l = [] then Format.fprintf fmt "." else aux true 0 l;
   Fmt.pp_close_tbox fmt ()
 ;;
 
-let pp_arg sep fmt (dir,ty,doc) =
-  let dir = if dir then "i" else "o" in
-  Fmt.fprintf fmt "%s:%s%s" dir ty sep
-;;
+let pp_args fmt l = 
+  let l1, l2 = List.partition (fun (x,_,_) -> x) l in
+  let pp_arg fmt (_,ty,_) = Format.fprintf fmt "%s" (parens_arr ty) in
+  let pp_args = pplist pp_arg ", " ~pplastelem:pp_arg in
+  if l1 <> [] then  Format.fprintf fmt " ";
+  Format.fprintf fmt "%a" pp_args l1;
+  if l2 = [] then ()
+  else if l1 = [] then Format.fprintf fmt " -> %a" pp_args l2
+  else Format.fprintf fmt " -> %a" pp_args l2
 
-let pp_args = pplist (pp_arg "") ", " ~pplastelem:(pp_arg "")
+let rec is_std_moded = function
+  | [] -> true
+  | (true,_,_) :: rest -> is_std_moded rest
+  | (false,_,_) :: [] -> true
+  | (false,_,_) :: ((false,_,_) :: _ as rest) -> is_std_moded rest
+  | _ -> false
 
+(* External predicate are functional *)
 let pp_pred fmt docspec name doc_pred args =
   let args = List.rev args in
-  match docspec with
-  | DocNext ->
-     Fmt.fprintf fmt "@[<v 2>external pred %s %% %s@;%a@]@."
-       name doc_pred pp_tab_args args
-  | DocAbove ->
-    let doc =
-       "[" ^ String.concat " " (name :: List.map (fun (_,_,x) -> x) args) ^
-       "] " ^ doc_pred in
-     Fmt.fprintf fmt "@[<v>%% %a@.external pred %s @[<hov>%a.@]@]@.@."
-       pp_comment doc name pp_args args
+  if is_std_moded args then
+    match docspec with
+    | DocNext ->
+      Fmt.fprintf fmt "@[<v 2>external func %s %% %s@;%a@]@."
+        name doc_pred pp_tab_args args
+    | DocAbove ->
+      let doc =
+        "[" ^ String.concat " " (name :: List.map (fun (_,_,x) -> x) args) ^
+        "] " ^ doc_pred in
+      Fmt.fprintf fmt "@[<v>%% %a@.external func %s@[<hov>%a.@]@]@.@."
+        pp_comment doc name pp_args args
+  else
+    let pp_tab_arg i max sep fmt (dir,ty,doc) =
+      let dir = if dir then "i" else "o" in
+      if i = 0 then Fmt.pp_set_tab fmt () else ();
+      Fmt.fprintf fmt "%s:%s%s" dir ty sep;
+      if i = 0 then (ws_to_max fmt max (String.length ty); Fmt.pp_set_tab fmt ()) else Fmt.pp_print_tab fmt ();
+      if doc <> "" then begin Fmt.fprintf fmt " %% %s" doc end;
+      Fmt.pp_print_tab fmt () in
+    let pp_tab_args fmt l =
+      let n = List.length l - 1 in
+      let max = List.fold_left (fun m (_,s,_) -> max (String.length s) m) 0 l in
+      Fmt.pp_open_tbox fmt ();
+      if l = [] then Fmt.fprintf fmt ".";
+      List.iteri (fun i x ->
+        let sep = if i = n then "." else "," in
+        pp_tab_arg i max sep fmt x) l;
+      Fmt.pp_close_tbox fmt () in
+    let pp_arg sep fmt (dir,ty,doc) =
+      let dir = if dir then "i" else "o" in
+      Fmt.fprintf fmt "%s:%s%s" dir ty sep in
+    let pp_args = pplist (pp_arg "") ", " ~pplastelem:(pp_arg "") in
+    match docspec with
+    | DocNext ->
+        Fmt.fprintf fmt "@\n@[<v 2>:functional :external pred %s %% %s@;%a@]@."
+          name doc_pred pp_tab_args args
+    | DocAbove ->
+      let doc =
+          "[" ^ String.concat " " (name :: List.map (fun (_,_,x) -> x) args) ^
+          "] " ^ doc_pred in
+        Fmt.fprintf fmt "@\n@[<v>%% %a@.:functional :external pred %s @[<hov>%a.@]@]@.@."
+          pp_comment doc name pp_args args   
 ;;
 
 let pp_variadictype fmt name doc_pred ty args =
-  let parens s = if String.contains s ' ' then "("^s^")" else s in
-  let args = List.rev ((false,"variadic " ^ parens ty ^ " prop","") :: args) in
+  let args = List.rev ((false,"variadic " ^ parens ty ^ " (func)","") :: args) in
   let doc =
     "[" ^ String.concat " " (name :: List.map (fun (_,_,x) -> x) args) ^
     "...] " ^ doc_pred in
@@ -1255,6 +1638,23 @@ let pp_variadictype fmt name doc_pred ty args =
         pp_comment doc name pp_ty_args args
 ;;
 
+let pp_variadicpred fmt docspec name doc_pred ty args =
+  let rargs = List.rev ((false, ty ^ ".. ","...") :: args) in
+  if is_std_moded rargs then
+    match docspec with
+    | DocNext ->
+      Fmt.fprintf fmt "@[<v 2>external func %s %% %s@;%a@]@."
+        name doc_pred pp_tab_args rargs
+    | DocAbove ->
+      let doc =
+        "[" ^ String.concat " " (name :: List.map (fun (_,_,x) -> x) rargs) ^
+        "] " ^ doc_pred in
+      Fmt.fprintf fmt "@[<v>%% %a@.external func %s@[<hov>%a.@]@]@.@."
+        pp_comment doc name pp_args rargs
+  else
+    pp_variadictype fmt name doc_pred ty args
+;;
+
 let document_pred fmt docspec name ffi =
   let rec doc
   : type i o h c. (bool * string * string) list -> (i,o,h,c) ffi -> unit
@@ -1270,8 +1670,8 @@ let document_pred fmt docspec name ffi =
     | Full (_,s) -> pp_pred fmt docspec name s args
     | FullHO (_,s) -> pp_pred fmt docspec name s args
     | VariadicIn( _,{ ContextualConversion.ty }, s) -> pp_variadictype fmt name s (Conversion.show_ty_ast ty) args
-    | VariadicOut( _,{ ContextualConversion.ty }, s) -> pp_variadictype fmt name s (Conversion.show_ty_ast ty) args
-    | VariadicInOut( _,{ ContextualConversion.ty }, s) -> pp_variadictype fmt name s (Conversion.show_ty_ast ty) args
+    | VariadicOut( _,{ ContextualConversion.ty }, s) -> pp_variadicpred fmt docspec name s (Conversion.show_ty_ast ty) args
+    | VariadicInOut( _,{ ContextualConversion.ty }, s) -> pp_variadicpred fmt docspec name s (Conversion.show_ty_ast ty) args
   in
     doc [] ffi
 ;;
@@ -1303,7 +1703,7 @@ type builtin_table = (int, t) Hashtbl.t
 end
 
 type symbol_table = {
-  mutable c2s : string Constants.Map.t;
+  mutable c2s : Symbol.t Constants.Map.t;
   c2t : (constant, term) Hashtbl.t;
   mutable frozen_constants : int;
 }
@@ -1341,4 +1741,4 @@ type solution = {
 }
 type 'a outcome = Success of solution | Failure | NoMoreSteps
 
-exception CannotDeclareClauseForBuiltin of Loc.t option * constant
+exception CannotDeclareClauseForBuiltin of Loc.t option * builtin_predicate
