@@ -95,6 +95,12 @@ let () = declare "chr_ut"
   ~description:"type checker with UT via CHR"
   ()
 
+let () = declare "chr_cut"
+  ~source_elpi:"chr_cut.elpi"
+  ~description:"cut not allowed"
+  ~expectation:(FailureOutput (Str.regexp "cannot contain cut"))
+  ()
+
 let () = declare "chr_even_odd"
   ~source_elpi:"even-odd.elpi"
   ~description:"CHR example at MLWS"
@@ -159,6 +165,20 @@ let () = declare "spill_collision"
   ~description:"spilling under 2 pi named the same"
   ()
 
+let () = declare "spill-var"
+  ~source_elpi:"spill-var.elpi"
+  ~description:"spilling a var"
+  ()
+
+let () = declare "spill2"
+  ~source_elpi:"spill2.elpi"
+  ~description:"spilling 2 arguments"
+  ()
+
+let () = declare "spill_sigma"
+  ~source_elpi:"spill_sigma.elpi"
+  ~description:"spilling under sigma"
+  ()
 
 let () = declare "block"
   ~source_elpi:"block.elpi"
@@ -271,9 +291,9 @@ let () = declare "printer"
 X0 is f X1 mod r X0
 X0 is f X1 + r X0 * g X2
 X0 is (f X1 + r X0) * g X2
-X0 is f X1 ^ r X0 ^ g X2
-X0 || X2 && X3 ===> X4
-[f X0, g X1, (a , b), a + b]
+X3 is f X1 ^ r X0 ^ g X2
+X0 || X2 && X4 ===> X5
+[f X0, g X1, a + b, a + b]
 |})))
   ()
 
@@ -489,4 +509,15 @@ let () = declare "impl_prec_silent"
 let () = declare "ifdef"
 ~source_elpi:"ifdef.elpi"
 ~description:"lexer ifdef"
-()
\ No newline at end of file
+()
+
+let () = declare "tc_ambiguous"
+  ~source_elpi:"tc_ambiguous.elpi"
+  ~description:"tc_ambiguous"
+  ~expectation:(FailureOutput Str.(regexp "too many"))
+  ()
+
+let () = declare "variadic"
+  ~source_elpi:"variadic.elpi"
+  ~description:"variadic"
+  ()
\ No newline at end of file
